#!/bin/sh
# \
exec wish "$0" "$@"

#----------------------------------------------------------------------
# samegame.tcl

# Copyright (c) 2005 Frank Bannon
# Started: Wed Jun 26 17:00:01 PDT 2002
# Copying: Distribution unlimited.  Please distribute.

# email: frankbannon@yahoo.com

# Please email me if you have any suggestions to improve this game.
# If anyone improves this code, please email the changes to me.

# Requires Tcl/Tk 8.0+.

# The Same Game is a simple tile-matching game, where you remove groups of
# matching pieces to receive points.
# The object is to earn the highest possible score.  You do receive a
# moderate bonus for removing all pieces.  

#----------------------------------------------------------------------

# MODIFICATION HISTORY
# for the full history of this program, read the "samegame.history" file.

# 1.15	2006.05.02
# added window icon

# 1.10	2005.05.03
# added sound support via snack, if available
# remove hangs for a second if remove.wav is missing

#----------------------------------------------------------------------

array set val {
 version	1.15
 revision	2006.05.02
 icon	same-game.ico
}

package require Tk
# snack needed for sound
catch {package require snack}

# default settings
array set game {
 rows		10
 cols		15
 theme		Default
 playername 	Player
 diff		5
 group		0
 groupNum	0
 score		0
 moves		0
 groups 	0
 time		0
 border		1
 minDiff	3
 maxDiff	7
 defDiff	5
 clearBonus	25
 sound		1
 highscoreNotify	0
 gamecount	0
 zoomed		0
 lastHighlight	0
 lastFilename	same.sav
 colors "black red yellow blue orange green purple brown"
 defcolors "black red yellow blue orange green purple brown"
 themes {
	Default {black red yellow blue orange green purple brown}
	Grayscale {black gray80 gray70 gray60 gray50 gray40 gray30 gray20}
	Camouflage {black #5f3d2f #5b5f2f #e0a85f #355b37 #7b573d #b9a47b #e0a85f}
	{Color Wheel} {black #0000fe #fe0000 #00fe00 #00fefe #fe00fe #fefe00 #d0d2da}
	Bold {black #5209b2 #ffdc16 #f82b04 #8615a7 #2e5a30 #2b7778 #00cd00}
	Pastel {black #ccfeff #89e69d #ffa8ff #765a4c #fcff7e #f7c292 #207eff}
	Winter {black #0000ca #d20000 #e07000 #00befe #006400 #8a008a #9eb0c8}
	}
}

array set sound {
 start	start.wav
 end		end.wav
 remove	remove.wav
 quit		quit.wav
 highscore	highscore.wav
}

# default key bindings: variable, text title, procedure to call, default key
set keydefs {
new	{New Game}	gameNew N
replay	{Replay Game}	gameReplay R
load	{Load Game}	gameLoad L
save	{Save Game}	gameSave A
undo1	{Undo}	undoMove Z
undo2	{Undo}	undoMove U
zoom	{Maximize Window}	zoomBoard M
count	{Count Pieces}	countShow C
scores	{High Scores}	scoresView H
sound	{Sound on/off}	toggleSound S
theme	{Next Theme}	themeNext T
keys	{Key Shortcuts}	bindKeySet K
}

# find icon
if {![file exists $val(icon)]} {
	set f [file join [file dir [info script]] $val(icon)]
	if {[file exists $f]} {set val(icon) $f}
	}

### Procedure:  bgerror

proc bgerror {message} {
tk_dialog .error Error "Error. $message" info 0 OK
}

### Procedure:  bindDefKeys
# bind all current keys to defaults

proc bindDefKeys {} {
global key keydefs
# unbind old keys first
foreach var [array names key] {
	unbindKey $key($var)
}
# bind all default keys
foreach {var txt prc def} $keydefs {
	set key($var) $def
	bindKey $def $var $prc
}
}

### Procedure:  bindKey
# bind one key to new value

proc unbindKey {oldkey} {
catch {event delete <<$event>> <KeyPress-$oldkey>} err
catch {bind . <KeyPress-[string toupper $oldkey]> ""}
catch {bind . <KeyPress-[string tolower $oldkey]> ""}
}

proc bindKey {newkey var prc} {
global key

# ignore known bad keys
switch $newkey {
	BackSpace	-
	Delete	-
	Tab	-
	Left	-
	Right	-
	Up	-
	Down	-
	Return	{
		return
		}
}

regsub { } $var {} event

if {[catch {event add <<$event>> <KeyPress-$newkey>} err]} {
	puts $err
	return
} else {
	# remove old binding
	unbindKey $key($var)
	set key($var) $newkey
}

catch {bind . <KeyPress-[string toupper $newkey]> $prc}
catch {bind . <KeyPress-[string tolower $newkey]> $prc}
}

### Procedure:  bindKeySet
# display a window to allow change of key bindings

proc bindKeySet {} {
global key keydefs
set f	.keys
if {[info command $f] != ""} {raise $f; return}
toplevel	$f
wm resizable	$f 1 1
wm minsize	$f 90 260
wm title	$f "Key Bindings"

frame	$f.fra1
frame	$f.fra2
frame	$f.fra3

label	$f.title -text "Key Bindings:"
pack	$f.title -in $f.fra1 -side left -anchor w

button $f.close -text " Close " -command "destroy $f"
balloon $f.close "Close this window"
pack $f.close -in $f.fra3 -side right

button $f.def -text " Defaults " -command {bindDefKeys}
balloon $f.def "Revert to default key bindings"
pack $f.def -in $f.fra3 -side left -padx 10


set x 0
foreach {var txt prc def} $keydefs {
	entry	$f.fra2.e$x -width 6 -textvariable key($var)
	label	$f.fra2.l$x -text "$txt" -anchor w
	bind	$f.fra2.e$x <Any-Key> \
		[namespace code "bindKey %K $var $prc; break"]
	grid	$f.fra2.e$x $f.fra2.l$x -sticky ew
	incr x
}

pack	$f.fra1 -side top -padx 2 -pady 1 -expand 0 -fill x
pack	$f.fra2 -side top -padx 2 -pady 1 -expand 0 -fill x
pack	$f.fra3 -side top -padx 2 -pady 1 -fill x
bind	$f <Return> "destroy $f"
bind	$f <Escape> "destroy $f"
}


### Procedure:  chooseColor
# select a color from a color selection window

proc chooseColor {num} {
global game
set color red
if {[info exists game(colors)]} {set color [lindex $game(colors) $num]}
set str [tk_chooseColor -initialcolor $color]
if {[string length $str]} {
	set game(colors) [lreplace $game(colors) $num $num $str]
	 .menu.game.colors entryconfigure $num -background $str
	updateBoard
	countShow update
} else {
	set str $color
}
return $str
}

### Procedure:  click
# user clicked on a game piece

proc click {args} {
global game board boardid
set id [.c find withtag current]
set num [lsearch $boardid $id]
if {[lindex $board $num]} {
	# piece clicked has value
	if {$game(groupNum) > 1} {
		# group of 2 or more clicked
		undoMove add
		incr game(score) [score $game(groupNum)]
		removePieces $game(group)
		compactPieces
		updateBoard
		set game(group) ""
		set game(groupNum) 0
		set game(lastHighlight) -1
		# count all remaining groups
		countGroups 1
	}
	status
	highlight
	if {$game(groups) < 1} {gameEnd}
}
}

### Procedure:  compactPieces
# compact all remaining pieces
# create list of a column's values, from bottom to top
# set all column values to zero
# go through list and copy all non-zero values into last empty column, bot to top

proc compactPieces {} {
global game board

# last fully empty column
set last_empty_c 1
for {set c 1} {$c <= $game(cols)} {incr c} {

set collist ""
for {set r $game(rows)} {$r >= 1} {incr r -1} {
	# create list of a column's values, from bottom to top
	lappend collist [getVal $r $c]
	# set all column values to zero
	setVal $r $c 0
}

# go through list and copy all non-zero values into last empty column, bot to top
set i $game(rows)
foreach x $collist {
	if {$x > 0} {
		setVal $i $last_empty_c $x
		incr i -1
	}
}

# check if last empty column is still empty
if {[getVal $game(rows) $last_empty_c] > 0} {incr last_empty_c}
} ;# end for c
}

### Procedure:  confirmEnd
# ask player if okay to quit game

proc confirmEnd {args} {
global game
if {$game(score) > 0} {
	set answer [tk_messageBox -title "Confirm Quit Game" \
	-message "Quit game?" -type yesno -icon question]
	switch -- $answer {
		yes	gameEnd
		no	return
	}
} else {
	gameEnd
}
}

### Procedure:  countGroups
# counts number of available groups
# returns on first group unless asked for total count

proc countGroups {{countall 0}} {
global game board
# total number of groups
set count 0
# list of all piece indices that are members of a group
set groups ""
set num -1
foreach value $board {
	incr num
	if {$value < 1} {continue}

	if {[lsearch $groups $num] > -1} {continue}
	# piece in not in another group

	set group [findGroup $num $value]

	if {[llength $group] > 1} {
		incr count
		foreach n $group {lappend groups $n}
		if {!$countall} {break}
		}
}
set game(groups) $count
return $count
}

### Procedure:  countPieces
# count the types of pieces on the board

proc countPieces {} {
global board game total
# initialize the piece count totals
for {set n 0} {$n <= $game(maxDiff)} {incr n} {set total($n) 0}
set total(total) 0
# count each piece, and total of all pieces
foreach n [lrange $board 1 end] {
	incr total($n)
	incr total(total)
}
}

### Procedure:  countShow
# counts total of each type of piece

proc countShow {{what 0}} {
global total game
set fra	.count
if {[info command $fra] != ""} {
	# window already exists
	raise $fra
	if {[string match $what update]} {
		# update piece colors
		for {set n 1} {$n <= $game(maxDiff)} {incr n} {
			$fra.f$n.c configure -bg [lindex $game(colors) $n]
		}		
	}
	return
}
if {[string match $what update]} {return}

toplevel	$fra
wm resizable	$fra 1 1
wm minsize	$fra 160 30
wm title	$fra "Count Pieces"
frame	$fra.fra3
frame	$fra.fra4

# for each piece color,
# create a vertical frame to hold a canvas above and a label below

for {set n 1} {$n <= $game(maxDiff)} {incr n} {
	frame	$fra.f$n
	set color [lindex $game(colors) $n]
	if {[string length $color] < 1} {set color black}
	canvas	$fra.f$n.c -bg $color -width 20 -height 20
	balloon	$fra.f$n.c "Piece type $n"
	label	$fra.f$n.t -textvariable total($n)
	balloon	$fra.f$n.t "Total pieces\nof type $n"

	bind	$fra.f$n.t <Double-Button-1> "gameSearch $n"

	pack	$fra.f$n.t -side bottom
	pack	$fra.f$n.c -side top -expand 1 -fill both
	pack	$fra.f$n	-in $fra.fra3 -side left -expand 1 -fill both
}

button	$fra.fra4.quit	-text " Close " -command "destroy $fra"
balloon	$fra.fra4.quit	"Close this window"
pack	$fra.fra4.quit	-side right

pack	$fra.fra3 -side top -padx 2 -pady 1 -expand 1 -fill both
pack	$fra.fra4 -side top -padx 2 -pady 1 -fill x
bind	$fra <Return> "destroy $fra"
bind	$fra <Escape> "destroy $fra"
}

### Procedure:  createBoard
# create the starting game board

proc createBoard {} {
global game board boardid total

set w [expr [winfo width  .c] / $game(cols)]
set h [expr [winfo height .c] / $game(rows)]
.c delete all
set boardid 0
for {set r 1} {$r <= $game(rows)} {incr r} {
	for {set c 1} {$c <= $game(cols)} {incr c} {
		set x1 [expr ($c-1)*$w]
		set y1 [expr ($r-1)*$h]
		set x2 [expr $c*$w]
		set y2 [expr $r*$h]
		lappend boardid [.c create rectangle $x1 $y1 $x2 $y2 \
			-fill black -tags {piece}]
	}
}
}

### Procedure:  findGroup
# find total pieces in group, and a list of group piece coords

proc findGroup {index {value 1} {group ""}} {
global game board

# check if this piece is already in the group, if so, return now
if {[lsearch $group $index] > -1} {return "$group"}

lappend group $index

# check up
set index2 $index
incr index2 -$game(cols)
if {[lindex $board $index2] == $value} {
	set group "[findGroup $index2 $value "$group"]"
}

# check down
set index2 $index
incr index2 $game(cols)
if {[lindex $board $index2] == $value} {
	set group "[findGroup $index2 $value "$group"]"
}

set colm [expr $index%$game(cols)]

# check right
if {$colm > 0} {
set index2 $index
incr index2
if {[lindex $board $index2] == $value} {
	set group "[findGroup $index2 $value "$group"]"
}}

# check left
if {$colm != 1} {
set index2 $index
incr index2 -1
if {[lindex $board $index2] == $value} {
	set group "[findGroup $index2 $value "$group"]"
}}

# set var with count of pieces in group
# return list of piece indexes in group
#set game(groupNum) [llength $group]
return "$group"
}

### Procedure:  gameEnd
# current game has no more groups

proc gameEnd {} {
global game board
incr game(gamecount)
if {[lindex $board [getIndex $game(rows) 1]] == 0} {
	# score bonus for removing all pieces
	set game(score) [expr int($game(score)*(1+($game(clearBonus)/100.0)))]
}
soundPlay end
if {$game(score) > 0} {scoresAdd}
gameNew
}

### Procedure:  gameExit
# exit the program

proc gameExit {} {
global val
soundPlay quit 1
settingsSave $val(filenameSettings)
exit
}

### Procedure:  gameLoad
# load a saved game

proc gameLoad {} {
global game board undo filetypes
if {![info exists game(lastFilename)]} {set game(lastFilename) same.sav}
set init_file [file tail $game(lastFilename)]
set init_dir [file dirname $game(lastFilename)]
set filename [tk_getOpenFile -initialdir $init_dir -initialfile $init_file \
	-filetypes $filetypes -title {Load Game}]
if {[string length $filename]} {
	set game(lastFilename) $filename

	# load the game data from file
	catch {source $filename}
	updateColors
	updateBoard 1
}
}

### Procedure:  gameNew
# start a new game

proc gameNew {} {
global game board
set board 0
set total(total) [expr $game(rows)*$game(cols)]
for {set x 1} {$x <= $total(total)} {incr x} {lappend board [random $game(diff)]}
set game(moves) 0
set game(score) 0
set game(groups) 0
countPieces
countGroups 1
undoMove clear
soundPlay start
status
updateBoard
}

## Procedure:  gameReplay
# replay the current game from the beginning

proc gameReplay {} {
undoMove replay
}

### Procedure:  gameSave
# save the current game to file

proc gameSave {} {
global val game board undo filetypes
if {![info exists game(lastFilename)]} {set game(lastFilename) same.sav}
set init_file [file tail $game(lastFilename)]
set init_dir [file dirname $game(lastFilename)]

set filename [tk_getSaveFile -initialdir $init_dir -initialfile $init_file \
	-filetypes $filetypes -title {Save Game}]
if {[string length $filename]} {
	set game(lastFilename) $filename

	# save the game data to file
	if {![catch {set fid [open $filename w 0644]} result]} {
		puts $fid "# $val(progtitle) v$val(version)  Saved Game"
		puts $fid "# Saved on [clock format [clock seconds] -format "%B %d %Y"]"
		puts $fid "\n"

		foreach {arry} {game} {
			foreach {variable} [lsort [array names $arry]] {
				set value [set ${arry}($variable)]
				if {[llength $value] > 1} {
					puts $fid "set ${arry}($variable) \"$value\""
				} else {
					puts $fid "set ${arry}($variable) $value"
				}
			}
		}
	puts $fid "set undo {$undo}"
	puts $fid "set board {$board}"
	close $fid
	} ;# end if catch
} ;# end string length filename
}

### Procedure:  gameSearch
# create a custom game board
# CHEAT

proc gameSearch {what} {
global game total
# reject search for piece type outside game difficulty
if {$what > $game(diff)} {return}
set newnum $total($what)

# ask for desired number of type what
set fra		.gamesearch
if {[info command $fra] != ""} {raise $fra; return}
toplevel	$fra
wm resizable	$fra 1 1
wm minsize	$fra 240 110
wm title	$fra "Choose New Amount"

frame	$fra.fra1 -background black
frame	$fra.fra2
frame	$fra.fra3

label	$fra.title -bg black -fg white -justify left \
	-text "\nChoose the desired amount\nof piece type $what\nRange is 2 to $total(total)\n"
pack	$fra.title	-in $fra.fra1 -side top -anchor center

label	$fra.lblname -text "Amount:"
pack	$fra.lblname	-in $fra.fra2 -side left -anchor w

entry	$fra.entry	-textvariable total(newnum)
pack	$fra.entry	-in $fra.fra2 -side left -expand 1 -fill x

button	$fra.close	-text " Ok " -command "gameSearch2 $what $fra"
balloon	$fra.close	"Create this game board"
pack	$fra.close	-in $fra.fra3 -side right

button	$fra.cancel	-text " Cancel " -command "destroy $fra"
pack	$fra.cancel	-in $fra.fra3 -side right -padx 10

pack	$fra.fra3 -side bottom -padx 2 -pady 1 -fill x
pack	$fra.fra1 -side top -padx 2 -pady 1 -fill x
pack	$fra.fra2 -side top -padx 2 -pady 1 -fill x
focus	$fra.entry
bind	$fra <Return> "gameSearch2 $what $fra"
bind	$fra <Escape> "destroy $fra"
}

### Procedure:  gameSearch2

proc gameSearch2 {what fra} {
global board total game
# check bounds
set newnum $total(newnum)
set top [llength $board]
if {($newnum < 2) || ($newnum >= $top)} {return}
if {$newnum > $total(total)} {return}
catch {destroy $fra}

# reduce pieces of given type
while {$total($what) > $newnum} {
	set x [random $top]
	set b [lindex $board $x]
	if {$b != $what} {continue}
	# find a replacement value for this piece
	set n 0
	while {($n < 1) || ($n == $what)} {set n [random $game(diff)]}
	incr total($n)
	incr total($what) -1
	set board [lreplace $board $x $x $n]
}

# add pieces of given type
# replace random pieces with desired type piece until number is reached
while {$total($what) < $newnum} {
	set x [random $top]
	set b [lindex $board $x]
	if {$b < 1} {continue}
	if {$b == $what} {continue}
	incr total($what)
	incr total($b) -1
	set board [lreplace $board $x $x $what]
}
updateBoard
}

### Procedure:  getWindowSize
# get current window sizes

proc getWindowSize {} {
global game
set wh [lindex [split [split [wm geometry .] x] +] 0]
set game(dezoomX) [lindex $wh 0]
set game(dezoomY) [lindex $wh 1]
}

### Procedure:  getRow
# return the row number given a list index number

proc getRow {num} {
global game
return [expr ($num-1)/$game(cols)+1]
}

### Procedure:  getCol
# return the column number given a list index number

proc getCol {num} {
global game
set col [expr $num%$game(cols)]
if {$col < 1} {set col $game(cols)}
return $col
}

### Procedure:  getIndex
# return the index given the row,col

proc getIndex {r c} {
global game
return [expr ($r-1)*$game(cols)+$c]
}

### Procedure:  getVal
# return the value at a given row,col spot
proc getVal {r c} {
global board
set num [lindex $board [getIndex $r $c]]
if {$num < 1} {set num 0}
return $num
}

### Procedure:  setVal
# set the value at a given row,col spot

proc setVal {r c value} {
global board
set num [getIndex $r $c]
set board [lreplace $board $num $num $value]
}

### Procedure:  highlight
# highlight current game piece

proc highlight {args} {
global game board boardid
set id [.c find withtag current]
set num [lsearch $boardid $id]
# remove highlight from last highlighted group
if {$num != $game(lastHighlight)} {
	foreach i $game(group) {
		updatePiece $i [lindex $game(colors) [lindex $board $i]]
	}
}
set r [getRow $num]
set c [getCol $num]
set value [lindex $board $num]

if {$value > 0} {
	set game(group) [findGroup $num $value]
	set game(groupNum) [llength $game(group)]
	if {$game(groupNum) > 1} {
		set game(lastHighlight) $num
		foreach i $game(group) {updatePiece $i white}
	}
} else {
	set game(groupNum) 0
}
status
update idletasks
}

### Procedure:  random
# generate a random number between 1 and num

proc random {num} {
return [expr int(rand()*$num+1)]
}

### Procedure:  removePieces
# remove all pieces in given group

proc removePieces {group} {
global game board total
incr game(moves)
set value [lindex $board [lindex $group 0]]
foreach n $group {
	# update totals of each piece type
	incr total($value) -1
	# remove piece
	set board [lreplace $board $n $n 0]
	soundPlay remove 1
	incr total(total) -1
}
}

### Procedure:  resizeBoard
# resize the game pieces to fit the new size game board

proc resizeBoard {} {
global game boardid
set w [winfo width  .c]
set h [winfo height .c]
if {![info exists game(sizex)]} {set game(sizex) $w; set game(sizey) $h}

# return if new size not enough to resize widgets
set dx [expr abs($game(sizex)-$w)]
set dy [expr abs($game(sizey)-$h)]
if {($dx < $game(cols)) && ($dy < $game(rows))} {return}

# save current size
set game(sizex) $w
set game(sizey) $h

set w [expr $w / $game(cols)]
set h [expr $h / $game(rows)]

for {set r 1} {$r <= $game(rows)} {incr r} {
for {set c 1} {$c <= $game(cols)} {incr c} {
	set x1 [expr ($c-1)*$w+1]
	set y1 [expr ($r-1)*$h+1]
	set x2 [expr $c*$w]
	set y2 [expr $r*$h]
	set id [lindex $boardid [expr [getIndex $r $c]]]
	.c coords $id $x1 $y1 $x2 $y2
} ;# end c
} ;# end r
}

### Procedure:  score
# compute score based on group size

proc score {num} {
if {$num < 2} {return 0}
return [expr ($num-2)*($num-2)+$num]
}

### Procedure:  scoresAdd
# add current score to high score list

proc scoresAdd {} {
global game highscore val
set s	$game(score)
set d	$game(diff)
if {[catch {exec date +%m.%d.%Y} result]} {
	set cdate "today"
} else {
	set cdate $result
}

if {($s < $highscore($d,10,score)) || ($s < 1)} {
	puts "   Score: $s ($game(moves) moves)  $game(playername)"
	return
}

# find position for new score
set pos 1
for {set x 1} {$x <= 10} {incr x} {
	if {$s < $highscore($d,$x,score)} {incr pos}
}

# shift old scores down
for {set x 10} {$x > $pos} {incr x -1} {
	set x1 [expr $x-1]
	set highscore($d,$x,score) $highscore($d,$x1,score)
	set highscore($d,$x,moves) $highscore($d,$x1,moves)
	set highscore($d,$x,name)  $highscore($d,$x1,name)
	set highscore($d,$x,date)  $highscore($d,$x1,date)
}

# place new score
set highscore($d,$pos,score) $s
set highscore($d,$pos,moves) $game(moves)
set highscore($d,$pos,name)  $game(playername)
set highscore($d,$pos,date)  $cdate

puts "High Score #$pos, Score: $s ($game(moves) moves)  $game(playername)"
scoresSave $val(filenameScores)
soundPlay highscore 1

# notify new high score if enabled
if {$game(highscoreNotify)} {
	tk_messageBox -icon info -type ok -title "High Score # $pos" \
		-message "Your score of $s takes #$pos on the high score table!"
}
}

### Procedure:  scoresLoad
# load high score table from file

proc scoresLoad {filename} {
global game highscore
if {![catch {set fid [open $filename r]} result]} {
# get two-line header info
set h1 [gets $fid]
set h2 [gets $fid]
for {set d 3} {$d <= $game(maxDiff)} {incr d} {
	for {set n 1} {$n <= 10} {incr n} {
		set temp [split [gets $fid] ,]
		set highscore($d,$n,score)	[lindex $temp 0]
		set highscore($d,$n,moves)	[lindex $temp 1]
		set highscore($d,$n,name)	[lindex $temp 2]
		set highscore($d,$n,date)	[lindex $temp 3]
	}
}
}

}

### Procedure:  scoresNew
# create a new high score list 

proc scoresNew {d} {
global highscore
for {set n 1} {$n <= 10} {incr n} {
	set highscore($d,$n,score)	0
	set highscore($d,$n,moves)	0
	set highscore($d,$n,name)	0
	set highscore($d,$n,date)	0
}
}

### Procedure:  scoresSave
# save the current high score list to file

proc scoresSave {filename} {
global game highscore val
if {[catch {set fid [open $filename w 0644]} result]} {
	showError "Cannot save high score table: $result"
} else {
	puts $fid "# $val(progtitle) v$val(version)  High Scores Table"
	puts $fid "# Saved on [clock format [clock seconds] -format "%B %d %Y"]"
	for {set d 3} {$d <= $game(maxDiff)} {incr d} {
		# verify a high score table exists
		# .ini file might have increased maxDiff
		if {![info exists highscore($d,1,score)]} {scoresNew $d}
		for {set n 1} {$n <= 10} {incr n} {
			set output [format "%s,%s,%s,%s" $highscore($d,$n,score) \
				$highscore($d,$n,moves) $highscore($d,$n,name) \
				$highscore($d,$n,date)]
			puts $fid $output
		}
	}
close $fid
}
}

### Procedure:  scoresView
# show the high score table

proc scoresView {} {
global game highscore
set fra .highscore
if {[info command $fra] != ""} {raise $fra; return}
set d $game(diff)
set hstable [format "\n%3s: %6s %6s  %-20s %-12s" N Score Moves Name Date]

for {set x 1} {$x <= 10} {incr x} {
	append hstable [format "\n%3s: %6s %6s  %-20s %-12s" $x \
	$highscore($d,$x,score) $highscore($d,$x,moves) \
	$highscore($d,$x,name)  $highscore($d,$x,date)]
}

toplevel $fra
wm resizable $fra 1 1
wm minsize $fra 300 200
wm title $fra "High Scores for difficulty $d"


frame	$fra.fra2 -background black
frame	$fra.fra3 -background white
frame	$fra.fra4

label $fra.fra2.lbl -bg black -fg white -font {helvetica 16 bold} \
	-text "High Scores for difficulty $d"
pack $fra.fra2.lbl -side top -anchor center

label	$fra.fra3.lbl -bg {#76A084} -anchor w -justify left \
	-text "$hstable" -font {courier 12 normal}
pack $fra.fra3.lbl -side top -anchor nw -expand 1 -fill both

button $fra.fra4.clear	-text "  Clear scores  " \
	-command "scoresNew $d; destroy $fra; scoresView"
balloon $fra.fra4.clear	"Clear these scores"
pack $fra.fra4.clear -side left

label $fra.fra4.total -text "Total games played:  $game(gamecount)"
pack $fra.fra4.total -padx 15 -side left

button $fra.fra4.quit -text "  Close  " -command "destroy $fra"
balloon $fra.fra4.quit "Close this window"
pack $fra.fra4.quit -side right

pack $fra.fra4 -side bottom -padx 2 -pady 1 -fill x
pack $fra.fra2 -side top -padx 2 -pady 1 -fill x
pack $fra.fra3 -side top -padx 2 -pady 1 -expand 1 -fill both
bind $fra <Return> "destroy $fra"
bind $fra <Escape> "destroy $fra"
}

### Procedure:  setPlayerName
# allow user to set the player name for high score entries

proc setPlayerName {} {
global game

set fra		.playername
if {[info command $fra] != ""} {raise $fra; return}
toplevel	$fra
wm resizable	$fra 1 1
wm minsize	$fra 240 110
wm title	$fra "Set Player Name"

frame	$fra.fra1 -background black
frame	$fra.fra2
frame	$fra.fra3

label	$fra.title -bg black -fg white \
	-text "\nPlease enter your player name\n"
pack	$fra.title	-in $fra.fra1 -side top -anchor center

label	$fra.lblname -text "Your Name:"
pack	$fra.lblname	-in $fra.fra2 -side left -anchor w

entry	$fra.entry	-textvariable game(playername)
pack	$fra.entry	-in $fra.fra2 -side left -expand 1 -fill x
set temp $game(playername)

button $fra.close	-text "  Close  " -command "destroy $fra"
balloon $fra.close	"Close this window"
pack $fra.close	-in $fra.fra3 -side right

pack	$fra.fra3 -side bottom -padx 2 -pady 1 -fill x
pack	$fra.fra1 -side top -padx 2 -pady 1 -fill x
pack	$fra.fra2 -side top -padx 2 -pady 1 -fill x
focus	$fra.entry
bind	$fra <Return> "destroy $fra"
bind	$fra <Escape> "destroy $fra"
}

### Procedure:  settingsLoad
# load saved settings

proc settingsLoad {filename} {
global game keydefs key
# unbind current keys
foreach {var txt prc def} $keydefs {unbindKey $key($var)}
catch {source $filename}
updateThemes
updateDifficulties
updateColors
# bind new keys
foreach {var txt prc def} $keydefs {bindKey $key($var) $var $prc}
}

### Procedure:  settingsSave
# save current settings to file

proc settingsSave {filename} {
global val game key
if {![catch {set fid [open $filename w 0644]} result]} {
puts $fid "# $val(progtitle) v$val(version)  Saved Variables"
puts $fid "# Saved on [clock format [clock seconds] -format "%B %d %Y"]"
puts $fid "\n"

foreach arry {game key} {
	foreach var [lsort [array names $arry]] {
		set value [set ${arry}($var)]
		if {[llength $value] > 1} {
			puts $fid "set ${arry}($var) \"$value\""
		} else {
			puts $fid "set ${arry}($var) $value"
		}
	}
}
close $fid
}
}

### Procedure:  showError
# display an error message

proc showError {message} {
tk_messageBox -default ok -type ok -icon error -title "Error" -message $message
}

### Procedure:  sortList
# sort a two-element list ascending or descending

proc sortList {dir lst} {
foreach {a b} $lst {lappend lst2 [list $a $b]}
if {[info exists lst2]} {
	switch -glob $dir {
		a* {set lst3 [lsort -index 0 -increasing $lst2]}
		d* {set lst3 [lsort -index 0 -decreasing $lst2]}
	}
	foreach {a b} $lst3 {append lst4 "$a $b "}
	return $lst4
} else {
	return $lst
}
}

### Procedure:  soundPlay
# play the selected sound

proc soundPlay {what {blocking 0}} {
global sound game
if {!$game(sound)} {return}
# check if sound was already loaded
if {![info exists snd_$what]} {
	catch {snack::sound snd_$what}
	if {[catch {snd_$what read $sound($what)}]} {catch {snd destroy}}
}
if {[catch {snd_$what play -blocking $blocking}]} {bell}
#catch {snack::sound snd}
#catch {snd read $sound($what)}
#if {[catch {snd play -blocking $blocking}]} {bell}
catch {snd destroy}
}

### Procedure:  status
# update the status bar with current group+score, total score, moves

proc status {} {
global game
if {[winfo exists .status]} {
set msg [format "Selected: %-4s   Moves: %-5s Total Score: %-6s Total Moves: %-5s" \
	[score $game(groupNum)] $game(groups) $game(score) $game(moves)]

	.status configure -text $msg
}
}

### Procedure:  themeAdd
# add the current color scheme as a named theme

proc themeAdd {} {
set fra	.theme
if {[info command $fra] != ""} {raise $fra; return}
toplevel	$fra
wm resizable	$fra 1 1
wm minsize	$fra 240 110
wm title	$fra "Add Theme"
set theme ""

frame	$fra.fra1 -background black
frame	$fra.fra2
frame	$fra.fra3

label	$fra.title -bg black -fg white \
	-text "\nPlease name the new color theme\n"
pack	$fra.title	-in $fra.fra1 -side top -anchor center

label	$fra.lblname -text "Theme Name:"
pack	$fra.lblname	-in $fra.fra2 -side left -anchor w

set theme ""
entry	$fra.entry	-textvariable theme
pack	$fra.entry	-in $fra.fra2 -side left -expand 1 -fill x
bind	$fra <Key-Return> "themeAdd2 $fra"

button	$fra.close	-text " Ok " -command "themeAdd2 $fra"
balloon	$fra.close	"Close this window"
pack	$fra.close	-in $fra.fra3 -side right

button	$fra.cancel	-text " Cancel " -command "destroy $fra"
balloon	$fra.cancel	"Cancel and do not add this theme"
pack	$fra.cancel	-in $fra.fra3 -side right -padx 10

pack	$fra.fra3 -side bottom -padx 2 -pady 1 -fill x
pack	$fra.fra1 -side top -padx 2 -pady 1 -fill x
pack	$fra.fra2 -side top -padx 2 -pady 1 -fill x
focus	$fra.entry
bind	$fra <Return> "$fra.close invoke"
bind	$fra <Escape> "destroy $fra"
}

### Procedure:  themeAdd2
# add the current color scheme as a named theme

proc themeAdd2 {fra} {
global game
set theme [$fra.entry get]
if {[string length $theme]} {
	set idx [lsearch $game(themes) $theme]
	lappend game(themes) $theme $game(colors)
	themeSet $theme
	if {$idx >= 0} {
		# theme already existed, so delete the old one
		themeDel
		themeSet $theme
	}
	updateThemes
}
destroy $fra
}

### Procedure:  themeDel
# delete the current theme, and rebuild the menu
# set the current theme to the first theme in the list

proc themeDel {} {
global game
if {[llength $game(themes)] > 2} {
	set idx [lsearch $game(themes) $game(theme)]
	set game(themes) [lreplace $game(themes) $idx [expr $idx +1]]
	themeSet [lindex $game(themes) 0]
	updateThemes
}
}

### Procedure:  themeNext
# set the current theme to next in list

proc themeNext {} {
global game
set idx [lsearch $game(themes) $game(theme)]
incr idx 2
if {$idx > [expr [llength $game(themes)]-1]} {set idx 0}
set newtheme [lindex $game(themes) $idx]
themeSet $newtheme
}

### Procedure:  themeSet
# set the current theme

proc themeSet {what} {
global game
set idx [lsearch $game(themes) $what]
if {$idx < 0} {set idx 0}
set game(theme) [lindex $game(themes) $idx]
incr idx
set game(colors) [lindex $game(themes) $idx]
set game(defcolors) $game(colors)
updateColors
}

### Procedure:  toggleSound
# support key binding to toggle sound

proc toggleSound {} {
global game
set game(sound) [expr ($game(sound)+1)%2]
}

### Procedure:  undoMove
# restore board to previous state

proc undoMove {{what del}} {
global game board undo
if {![info exists undo]} {set undo ""}
switch -glob -- $what {
add*	{
	lappend undo [list $game(score) $game(moves) $game(time) $game(groups) \
		$board]
	}
del*	{
	if {[llength $undo]} {
		set str [lindex $undo end]
		set game(score) [lindex $str 0]
		set game(moves) [lindex $str 1]
		set game(time) [lindex $str 2]
		set game(groups) [lindex $str 3]
		set game(groupNum) 0
		set board [lindex $str end]
		set undo [lreplace $undo end end]
		updateBoard
		status
		highlight
		}
	}
re*	{
	if {[llength $undo]} {
		set str [lindex $undo 0]
		set game(score) [lindex $str 0]
		set game(moves) [lindex $str 1]
		set game(time) [lindex $str 2]
		set game(groups) [lindex $str 3]
		set game(groupNum) 0
		set board [lindex $str end]
		set undo ""
		updateBoard
		status
		highlight
		}
	}
cl*	{
	set undo ""
	}
}
}

### Procedure:  updateBoard
# update entire game board
# update all pieces, but check each piece color and update only if necessary
# for much better full-screen performance

proc updateBoard {{force 0}} {
global game board boardid
if {![info exists board]} {return}

for {set i 0} {$i < [llength $board]} {incr i} {
	# read correct color for piece
	set new_color [lindex $game(colors) [lindex $board $i]]
	# read current piece color
	set color [lrange [.c itemconfigure [lindex $boardid $i] -fill] end end]
	if {$force} {
		updatePiece [lindex $boardid $i] $new_color
	} else {
		if {![string match $color $new_color]} {
			 updatePiece [lindex $boardid $i] $new_color
		}
	}
}
}

### Procedure:  updateColors
# update the colros menu with new entries

proc updateColors {args} {
global game
# delete all items on the menu
catch {.menu.game.colors delete 0 end}
# check that colors list has all necessary entries
set c1 [llength $game(colors)]
if {$game(maxDiff) >= $c1} {
	for {set x 1} {$x <= [expr $game(maxDiff)-$c1+1]} {incr x} {
		lappend game(colors) [lindex $game(colors) end]
	}
}

for {set x 1} {$x <= $game(maxDiff)} {incr x} {
	.menu.game.colors add command -label $x -underline 0 \
	-background [lindex $game(colors) $x] -command "chooseColor $x"
}
.menu.game.colors add separator
.menu.game.colors add command -label {Default colors} -underline 0 \
	-command {set game(colors) $game(defcolors); updateColors}
countShow update
updateBoard 1
}

### Procedure:  updateDifficulties
# update the difficulties menu with new entries

proc updateDifficulties {args} {
global game
# delete all items on the menu
catch {.menu.game.diff delete 0 end}
for {set x $game(minDiff)} {$x <= $game(maxDiff)} {incr x} {
	.menu.game.diff add check -label $x -underline 0 -variable game(diff) \
		-onvalue $x -command "gameNew"
}
.menu.game.diff add separator
.menu.game.diff add command -label {Default difficulty} -underline 0 \
	-command "set game(diff) $game(defDiff); gameNew"
}

### Procedure:  updatePiece
# change state of a piece

proc updatePiece {num {color black}} {
global game
.c itemconfigure $num -fill $color -width $game(border)
}

### Procedure:  updateThemes
# update the theme menu with new entries

proc updateThemes {args} {
global game
# delete all themes on the menu
catch {.menu.game.theme delete 0 end}
set game(themes) [sortList a $game(themes)]
foreach {n x} $game(themes) {
 .menu.game.theme add check -label $n -underline 0 -variable game(theme) \
	-onvalue $n -command "themeSet {$n}"
}
.menu.game.theme add separator
.menu.game.theme add command -label {Save New Theme ...} -underline 0 \
	-command {themeAdd}
.menu.game.theme add command -label {Delete Current Theme} -underline 0 \
	-command {themeDel}
# set current theme
themeSet $game(theme)
}

### Procedure:  zoomBoard
# zoom/dezoom the game board

proc zoomBoard {} {
global game
if {$game(zoomed)} {
	# dezoom
	wm geometry	. $game(dezoomX)x$game(dezoomY)
	set game(zoomed) 0
} else {
	# zoom to max size
	foreach {X Y} [wm maxsize .] {}
	incr Y -60
	getWindowSize
	wm geometry	. ${X}x$Y
	resizeBoard
	set game(zoomed) 1
}
}

### Procedure:  about

proc about {} {
variable val
variable game
if {[info command .about] != ""} {raise .about; return}
toplevel	.about
wm resizable	.about 0 0
wm minsize	.about 400 200
wm title	.about "About this program"

frame	.about.fra1 -background white
frame	.about.fra2 -background black
frame	.about.fra3	-background {#76A084}
frame	.about.fra4

if {[info exists val(filenameLogo)]} {
	set logofilename $val(filenameLogo)
	set temp [file tail $val(filenameLogo)]
	if {[file readable $temp]} {set logofilename $temp}
	if {[file readable $logofilename]} {
		set logo [image create photo -file $logofilename]
		label	.about.fra1.logo -image $logo
		pack	.about.fra1.logo -side top
	}
}

label	.about.fra2.lbl -bg black -fg white -font {helvetica 16 bold} \
	-text "$val(progtitle) $val(version)"
pack	.about.fra2.lbl		-side top -anchor center

set about_text ""
foreach {n t} {
copyyear	"Copyright (C) %s"
company		"  by %s"
release		"\nReleased as %s"
version		"\n\nVersion:   %s"
revision	"\nRevised:  %s"
author		"\n\nAuthor:   %s"
authorEmail	"\nEmail:     %s"
website		"\nWeb site: %s"
about		"\n\n%s\n"
} {
if {[info exists val($n)]} {append about_text [format $t $val($n)]}
}

label	.about.fra3.lbl -bg {#76A084} -anchor w -justify left -wrap 390 \
	-text "$about_text"
pack	.about.fra3.lbl		-side top -anchor nw -expand 1 -fill both

if {[info exists val(filenameHistory)]} {
if {[file readable $val(filenameHistory)]} {
	button	.about.fra4.vers	-text "Revisions" \
		-command {showFile $val(filenameHistory)}
	balloon	.about.fra4.vers	"Read the revision log for this program"
pack	.about.fra4.vers	-side left
}}

if {[info exists val(filenameReadme)]} {
	if {[file readable $val(filenameReadme)]} {
		button	.about.fra4.readme	-text "Read Me" \
			-command {showFile $val(filenameReadme)}
		balloon	.about.fra4.readme	"Read the readme file for this program"
		pack	.about.fra4.readme	-side left
	}
}

button	.about.fra4.quit	-text "  Close  " -command {destroy .about}
balloon	.about.fra4.quit	"Close this window"
pack	.about.fra4.quit	-side right

pack	.about.fra4 -side bottom -padx 2 -pady 1 -fill x
pack	.about.fra1 -side top -padx 2 -pady 1 -fill x
pack	.about.fra2 -side top -padx 2 -pady 1 -fill x
pack	.about.fra3 -side top -padx 2 -pady 1 -expand 1 -fill both
bind	.about <Return> "destroy .about"
bind	.about <Escape> "destroy .about"
}

### Procedure:  balloon
# the two balloon procedures came from http://mini.net/tcl/
proc balloon {w help {help2 ""}} {
global val
# tooltips give error on mac
if {$::tcl_platform(platform) == "macintosh"} {return}
if {[string length $help2] < 1} {set help2 $help}
bind $w <Any-Enter> "after $val(balloonDelay) [list balloon:show %W [list $help]]"
bind $w <Any-Leave> "destroy %W.balloon"
}

proc balloon:show {w arg} {
if {[eval winfo containing  [winfo pointerxy .]]!=$w} {return}
set top $w.balloon
catch {destroy $top}
toplevel $top -bd 1 -bg gray
wm overrideredirect $top 1
pack [message $top.txt -aspect 10000 -bg lemonchiffon1 -font fixed -width 170 \
	-text $arg]
set wmx [winfo rootx $w]
set wmy [expr [winfo rooty $w]+[winfo height $w]]
wm geometry $top [winfo reqwidth $top.txt]x[winfo reqheight $top.txt]+$wmx+$wmy
raise $top
}

### Procedure:  showFile
# displays a text file

proc showFile {filename} {
global val
set fra	.revisions
if {![file readable $filename]} {return}
if {[info command $fra] != ""} {raise $fra; return}
toplevel	$fra
wm geometry	$fra 450x300; update
wm minsize	$fra 300 200
wm resizable	$fra 1 1
wm title	$fra "$filename"

text	$fra.text -background white -yscrollcommand "$fra.scr set" \
	-width 30 -wrap word
scrollbar $fra.scr -command "$fra.text yview"

if {[catch {set fid [open $filename r]} result]} {return}
while {[gets $fid line] != -1} {$fra.text insert end $line\n}
close $fid

frame	$fra.b
button	$fra.b.close -text "  Close  " -command "destroy $fra"
balloon	$fra.b.close "Close this window"
pack	$fra.b.close -side right -padx 5 -expand 1 -anchor e
pack	$fra.b -side bottom -fill x
pack	$fra.scr -side right -fill y
pack	$fra.text -side left -expand true -fill both
focus	$fra.text
bind	$fra <Return> "destroy $fra"
bind	$fra <Escape> "destroy $fra"
}

#----------------------------------------------------------------
### Procedure:  main

proc main {argc argv} {
global val game
# create new high score table
for {set d 3} {$d <= $game(maxDiff)} {incr d} {scoresNew $d}
update
bindDefKeys
settingsLoad $val(filenameSettings)
scoresLoad $val(filenameScores)
createBoard
gameNew

update idletasks
getWindowSize
bind . <Configure> {resizeBoard}
}

### Procedure:  init

proc init {argc argv} {
global val filetypes

# non-user-changeable constants, stored in a convenient array
array set val {
 progtitle	"Same Game"
 company	"Frank Bannon"
 author		"Frank Bannon"
 authorEmail	"frankbannon@yahoo.com"
 website	"http://www.geocities.com/frankbannon/same/"
 copyyear	2006
 filenameReadme	"readme.txt"
 filenameHistory	"history.txt"
 filenameHowToPlay	"how-to-play.txt"
 filenameKeys		"key-maps.txt"
 filenameLogo		"samegame.gif"
 filenameSettings	"samegame.ini"
 filenameScores		"samegame.scores"
 release		"Copyrighted Freeware"
 balloonDelay		1200
 about	"A simple game of removing groups of adjacent matching pieces.  You receive points based on the size of each group removed.  Larger groups earn you more points.  The object of the game is to earn the highest possible score.  Removal of all pieces will add a moderate bonus to your final score."
}

# save game filetypes
set filetypes {
{{Saved Games}	.sav	}
{{All Files}	*	}
}

}

init $argc $argv

wm geometry . 426x350; update
wm minsize . 140 140
wm resizable . 1 1
wm title . "$val(progtitle)  v$val(version)"
catch {wm iconbitmap . -default $val(icon)}

. configure -menu .menu

menu .menu -tearoff 1
.menu add cascade -menu .menu.file		-label File -underline 0
.menu add cascade -menu .menu.game		-label Game -underline 0
.menu add cascade -menu .menu.help		-label Help -underline 0

menu .menu.file -tearoff 0
.menu.file add command -label {Load Game...} -command {gameLoad} -underline 0
.menu.file add command -label {Save Game...} -command {gameSave} -underline 0
.menu.file add command -label {Replay Game} -command {undoMove replay} -underline 0
.menu.file add separator
.menu.file add command -label {Exit} -command {gameExit} -underline 1

menu .menu.game -tearoff 1
.menu.game add command -label {Undo} -command {undoMove del} -underline 0
.menu.game add separator
.menu.game add command -label {New Game} -command {confirmEnd} -underline 0
.menu.game add separator

.menu.game add command -label "High Scores..." -command {scoresView} -underline 0
.menu.game add command -label "Count Pieces..." -command {countShow} -underline 0
.menu.game add separator

# difficulty
.menu.game add cascade -menu .menu.game.diff -label {Difficulty} -underline 0
menu .menu.game.diff -tearoff 1
balloon .menu.game.diff "Set the number of different pieces on the board"
updateDifficulties

# colors
.menu.game add cascade -menu .menu.game.colors -label {Colors} -underline 2
menu .menu.game.colors -tearoff 1
balloon .menu.game.colors "Change the game board colors"
updateColors

# themes
.menu.game add cascade -menu .menu.game.theme -label {Themes} -underline 0
menu .menu.game.theme -tearoff 1
balloon .menu.game.theme "Apply, save, or delete a set of colors"
updateThemes

# border size
.menu.game add cascade -menu .menu.game.border -label {Border size} -underline 0
menu .menu.game.border -tearoff 1
balloon .menu.game.border "Set the border size of the pieces"
for {set x 0} {$x <= 6} {incr x} {
 .menu.game.border add check -label $x -underline 0 -variable game(border) \
	-onvalue $x -command {updateBoard 1}
}

.menu.game add separator
.menu.game add checkbutton -label "Use Sound" -underline 4 -variable game(sound)
.menu.game add checkbutton -label "Alert high score" -underline 0 \
	-variable game(highscoreNotify)

.menu.game add separator
.menu.game add command -label {Key shortcuts...} -command bindKeySet -underline 0
.menu.game add command -label "Set Player Name..." -underline 1 \
	-command {setPlayerName}

menu .menu.help -tearoff 0
if {[info exists val(filenameHowToPlay)]} {
	if {[file readable $val(filenameHowToPlay)]} {
		.menu.help add command -label {How to Play...} \
		-command {showFile $val(filenameHowToPlay)} -underline 0
		.menu.help add separator
	}
}
.menu.help add command -label {About...} -command {about} -underline 0


label .status -relief ridge -anchor w
pack .status -side bottom -fill x

canvas .c -bg black
pack .c -side top -fill both -expand 1


.c bind piece <Enter> "highlight"
.c bind piece <1> "click"
bind . <3> {undoMove del}

main $argc $argv

# end SAME GAME 1.10 in TCL/TK, Copyright 2005 by Frank Bannon
